from dataclasses import dataclass

from cpm.domain.template_installer import TemplateInstaller
from cpm.infrastructure import filesystem
from cpm.domain.project.project import Project
from cpm.domain.project.project_descriptor_parser import ProjectDescriptorNotFound
from cpm.domain.project import project_descriptor_editor
from cpm.domain.project.project_loader import ProjectLoader
from cpm.infrastructure.cpm_hub_connector_v1 import CpmHubConnectorV1
from cpm.domain.sample_code import CPP_HELLO_WORLD
from cpm.domain.constants import PROJECT_DESCRIPTOR_FILE, INITIAL_PROJECT_VERSION


@dataclass
class CreationOptions:
    project_name: str = 'ProjectName'
    directory: str = '.'
    repository_url: str = 'https://repo.cpmbits.com:8000'
    init_from_existing_sources: bool = False
    generate_sample_code: bool = True
    init_from_template: bool = False
    template_name: str = ''
    template_version: str = 'latest'


class CreationService:
    def __init__(self, project_loader=ProjectLoader(), cpm_hub_connector=CpmHubConnectorV1(), template_installer=TemplateInstaller()):
        self.template_installer = template_installer
        self.project_loader = project_loader
        self.cpm_hub_connector = cpm_hub_connector

    def exists(self, directory):
        try:
            self.project_loader.load(directory)
            return True
        except ProjectDescriptorNotFound:
            return False

    def create(self, options):
        if options.init_from_template:
            return self.__create_from_template(options)
        elif options.init_from_existing_sources:
            return self.__create_from_existing_sources(options)
        else:
            return self.__create_from_scratch(options)

    def __create_from_template(self, options):
        template_download = self.cpm_hub_connector.download_template(options.template_name, options.template_version)
        self.template_installer.install(template_download, options.directory)
        project = self.project_loader.load(options.directory)
        project_descriptor_editor.update(
            options.directory,
            project.descriptor,
            {
                'name': options.project_name,
                'version': INITIAL_PROJECT_VERSION,
            }
        )
        return project

    def __create_from_existing_sources(self, options):
        project = Project(options.project_name)
        self.create_project_descriptor_file(options)
        if options.generate_sample_code:
            self.generate_sample_code(project)
        return project

    def __create_from_scratch(self, options):
        project = Project(options.project_name)

        self.create_project_directory(options.directory)
        filesystem.create_directory(f'{options.directory}/tests')
        self.create_project_descriptor_file(options)
        if options.generate_sample_code:
            self.generate_sample_code(project)
        return project

    def generate_sample_code(self, project):
        filesystem.create_file(
            f'{project.name}/main.cpp',
            CPP_HELLO_WORLD
        )

    def create_project_descriptor_file(self, options):
        filesystem.create_file(
            f"{options.directory}/{PROJECT_DESCRIPTOR_FILE}",
            f"name: '{options.project_name}'\n"
            f"version: {INITIAL_PROJECT_VERSION}\n"
            f"build:\n"
            f"  packages:\n"
            f"  bits:\n"
            f"test:\n"
            f"  bits:\n"
            f"targets:\n"
            f"  default:\n"
            f"    main: 'main.cpp'\n"
        )

    def create_project_directory(self, project_name):
        filesystem.create_directory(project_name)



import unittest
import mock

from cpm.domain.project.project import Project
from cpm.domain.project.project_descriptor_parser import ProjectDescriptorNotFound
from cpm.domain.test_service import TestService, NoTestsFound


class TestTestService(unittest.TestCase):
    def setUp(self):
        self.cmakelists_builder = mock.MagicMock()
        self.project_loader = mock.MagicMock()
        self.project_commands = mock.MagicMock()
        self.test_service = TestService(self.project_loader, self.cmakelists_builder, self.project_commands)

    def test_service_fails_when_project_loader_fails_to_load_project(self):
        self.project_loader.load.side_effect = ProjectDescriptorNotFound

        self.assertRaises(ProjectDescriptorNotFound, self.test_service.run_tests, [], 'default')

        self.project_loader.load.assert_called_once()

    def test_service_raises_exception_when_no_test_suites_are_found(self):
        project = Project('ProjectName')
        self.project_loader.load.return_value = project

        self.assertRaises(NoTestsFound, self.test_service.run_tests, [], 'default')

        self.project_loader.load.assert_called_once()

    def test_service_generates_the_recipe_then_compiles_and_runs_the_tests(self):
        project = Project('ProjectName')
        project.test.test_suites = ['test']
        self.project_loader.load.return_value = project

        self.test_service.run_tests([], 'default')

        self.project_loader.load.assert_called_once()
        self.cmakelists_builder.build.assert_called_once_with(project)
        self.project_commands.build_tests.assert_called_once_with(project, [])
        self.project_commands.run_tests.assert_called_once_with(project, [], ())

    def test_service_build_and_runs_only_specified_tests(self):
        project = Project('ProjectName')
        project.test.test_suites = ['test']
        self.project_loader.load.return_value = project

        self.test_service.run_tests(['tests'], 'default')

        self.project_loader.load.assert_called_once()
        self.cmakelists_builder.build.assert_called_once_with(project)
        self.project_commands.build_tests.assert_called_once_with(project, ['tests'])
        self.project_commands.run_tests.assert_called_once_with(project, ['tests'], ())

    def test_service_passes_test_args_to_run_command(self):
        project = Project('ProjectName')
        project.test.test_suites = ['test']
        self.project_loader.load.return_value = project

        self.test_service.run_tests(['tests'], 'default', test_args=['arg1'])

        self.project_loader.load.assert_called_once()
        self.cmakelists_builder.build.assert_called_once_with(project)
        self.project_commands.build_tests.assert_called_once_with(project, ['tests'])
        self.project_commands.run_tests.assert_called_once_with(project, ['tests'], ['arg1'])



import unittest
from mock import MagicMock

from cpm.api.clean import clean_project
from cpm.domain.project.project_descriptor_parser import ProjectDescriptorNotFound


class TestApiClean(unittest.TestCase):
    def test_clean_fails_when_current_directory_is_not_a_cpm_project(self):
        compilation_service = MagicMock()
        compilation_service.clean.side_effect = ProjectDescriptorNotFound()

        result = clean_project(compilation_service)

        assert result.status_code == 1
        compilation_service.clean.assert_called_once()

    def test_clean_project(self):
        compilation_service = MagicMock()

        result = clean_project(compilation_service)

        assert result.status_code == 0
        compilation_service.clean.assert_called_once()



from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

from cpm.domain.constants import PROJECT_DESCRIPTOR_FILE
from cpm.infrastructure import filesystem


def update(directory, project_descriptor, params):
    stream = StringIO()
    project_descriptor.yaml_document.update(params)
    yaml = YAML()
    yaml.dump(project_descriptor.yaml_document, stream=stream)
    yaml_payload = stream.getvalue()
    filesystem.write_file(f'{directory}/{PROJECT_DESCRIPTOR_FILE}', yaml_payload)



import unittest

from cpm.domain.project import project_descriptor_parser
from cpm.domain.project.project_descriptor import DeclaredBit, CompilationPlan, PackageDescription


class TestCompilationPlanParser(unittest.TestCase):
    def test_parsing_empty_compilation_plan(self):
        plan_description = {}
        compilation_plan = project_descriptor_parser.parse_compilation_plan(plan_description)
        assert compilation_plan == CompilationPlan()

    def test_parsing_compilation_plan_with_declared_bits(self):
        plan_description = {
            'bits': {
                'sqlite3': '1.0',
                'base64': '3.22'
            }
        }
        compilation_plan = project_descriptor_parser.parse_compilation_plan(plan_description)
        assert compilation_plan.declared_bits == [DeclaredBit('sqlite3', '1.0'), DeclaredBit('base64', '3.22')]

    def test_parsing_compilation_plan_with_declared_packages(self):
        plan_description = {
            'packages': {
                'cpmhub/bits': {},
                'cpmhub/http': {}
            }
        }
        compilation_plan = project_descriptor_parser.parse_compilation_plan(plan_description)
        assert compilation_plan.packages == [PackageDescription('cpmhub/bits'), PackageDescription('cpmhub/http')]

    def test_parsing_compilation_plan_with_declared_packages_with_cflags(self):
        plan_description = {
            'packages': {
                'cpmhub/bits': {
                    'cflags': ['-O2']
                },
            }
        }
        compilation_plan = project_descriptor_parser.parse_compilation_plan(plan_description)
        assert compilation_plan.packages == [PackageDescription('cpmhub/bits', cflags=['-O2'])]

    def test_parsing_compilation_plan_with_flags_and_libraries(self):
        plan_description = {
            'cflags': ['-O0'],
            'libraries': ['pthread']
        }
        compilation_plan = project_descriptor_parser.parse_compilation_plan(plan_description)
        assert compilation_plan.cflags == ['-O0']
        assert compilation_plan.libraries == ['pthread']




from dataclasses import dataclass
from urllib3.exceptions import InsecureRequestWarning
import requests


requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

@dataclass
class HttpResponse:
    status_code: int
    body: str


def post(url, data=None, headers=None, files=None):
    try:
        response = requests.post(url, files=files, data=data, headers=headers, verify=False)
        return HttpResponse(response.status_code, response.text)
    except requests.exceptions.ConnectionError as e:
        raise HttpConnectionError()


def put(url, data=None, headers=None, files=None):
    requests.put(url, files=files, data=data, headers=headers, verify=False)


def get(url, data=None, headers=None, files=None):
    try:
        response = requests.get(url, files=files, data=data, headers=headers, verify=False)
        return HttpResponse(response.status_code, response.text)
    except requests.exceptions.ConnectionError:
        raise HttpConnectionError(url)


class HttpConnectionError(RuntimeError):
    pass



import unittest
import mock

from cpm.api.init import init_project
from cpm.domain.creation_service import CreationOptions


class TestApiInit(unittest.TestCase):
    def test_init_project_uses_creation_service_for_initializing_project(self):
        creation_service = mock.MagicMock()
        creation_service.exists.return_value = False
        options = CreationOptions(
            init_from_existing_sources=True,
            directory='.',
            project_name='Awesome Project'
        )

        result = init_project(creation_service, options)

        assert result.status_code == 0
        creation_service.create.assert_called_once_with(options)

    def test_returns_error_code_when_project_already_exists(self):
        creation_service = mock.MagicMock()
        creation_service.exists.return_value = True

        result = init_project(creation_service)

        assert result.status_code == 1



class CMakeListsBuilder(object):
    def __init__(self):
        self.contents = ''
        self.object_libraries = []
        self.test_object_libraries = []

    def build(self, project):
        self.build_contents(project)
        with open('CMakeLists.txt', 'w') as cmakelists_file:
            cmakelists_file.write(self.contents)

    def build_contents(self, project):
        self.minimum_required('3.13')
        if project.target.toolchain_prefix:
            self.set_compilers(project.target.toolchain_prefix)
        self.project(project.name)
        for package in self.bit_packages_with_sources(project.target):
            self.object_libraries += self.build_package_recipe(package)
        for package in self.target_packages_with_sources(project.target):
            self.object_libraries += self.build_package_recipe(package)
        self.link_libraries(project.target.libraries)
        self.add_executable(
            project.name,
            [project.target.main],
            self.object_libraries
        )
        self.set_target_properties(project.name, 'COMPILE_FLAGS', self.main_compile_flags(project, project.target.main))
        self.target_link_options(project.name, project.target.ldflags)
        self.include_directories(project.target.include_directories)
        for package in self.test_packages_with_sources(project.test):
            self.test_object_libraries += self.build_package_recipe(package)
        for package in self.bit_packages_with_sources(project.test):
            self.test_object_libraries += self.build_package_recipe(package)
        for test in project.test.test_suites:
            self.add_executable(
                test.name,
                [test.main],
                self.object_libraries + self.test_object_libraries
            )
            self.set_target_properties(test.name, 'COMPILE_FLAGS', self.test_main_compile_flags(project, test.main))
            self.target_link_options(test.name, project.test.ldflags)
            self.target_include_directories(
                test.name,
                project.test.include_directories
                    .union(test.include_directories)
            )
            self.target_link_libraries(test.name, project.test.libraries + test.libraries)
        if project.test.test_suites:
            self.add_custom_target('tests', 'echo ""', [test.name for test in project.test.test_suites])
        return self.contents

    def target_packages_with_sources(self, target):
        return [package for package in target.packages if package.sources]

    def bit_packages_with_sources(self, target):
        return [package for bit in target.bits for package in bit.packages if package.sources]

    def test_packages_with_sources(self, test):
        return [package for package in test.packages if package.sources]

    def build_package_recipe(self, package):
        package_c_library_name = self.object_library_name(package.path + '_c')
        package_cpp_library_name = self.object_library_name(package.path + '_cpp')
        return self.build_package_library_recipe(package, package_c_library_name, package.cflags, '.c') + \
               self.build_package_library_recipe(package, package_cpp_library_name, package.cppflags, '.cpp')

    def build_package_library_recipe(self, package, package_library_name, compile_flags, extension):
        sources = list(filter(lambda s: s.endswith(extension), package.sources))
        if not sources:
            return []
        self.add_object_library(package_library_name, sources)
        self.set_target_properties(package_library_name, 'COMPILE_FLAGS', compile_flags)
        self.target_include_directories(package_library_name, package.include_directories)
        return [f'$<TARGET_OBJECTS:{package_library_name}>']

    def object_library_name(self, package_path):
        return f'{package_path.replace("/", "_")}_object_library'

    def minimum_required(self, version):
        self.contents += f'cmake_minimum_required (VERSION {version})\n'

    def project(self, name):
        self.contents += f'project({name})\n'

    def include_directories(self, directories):
        self.contents += f'include_directories({" ".join(sorted(directories))})\n'

    def set_source_files_properties(self, sources, property, values):
        self.contents += f'set_source_files_properties({" ".join(sources)} PROPERTIES {property} "{" ".join(values)}")\n'

    def add_object_library(self, name, sources):
        self.contents += f'add_library({name} OBJECT {" ".join(sources)})\n'

    def add_static_library(self, name, sources):
        self.contents += f'add_library({name} STATIC {" ".join(sources)})\n'

    def add_executable(self, name, sources, object_libraries=[]):
        self.contents += f'add_executable({name} {" ".join(sources)}'
        for library in object_libraries:
            self.contents += f' {library}'
        self.contents += ')\n'

    def main_compile_flags(self, project, main_filename):
        if main_filename.endswith('.c'):
            return project.target.cflags
        elif main_filename.endswith('.cpp'):
            return project.target.cppflags
        return []

    def test_main_compile_flags(self, project, main_filename):
        if main_filename.endswith('.c'):
            return project.target.cflags + project.test.cflags
        elif main_filename.endswith('.cpp'):
            return project.target.cppflags + project.test.cppflags
        return []

    def set_target_properties(self, target, property, values):
        if values:
            self.contents += f'set_target_properties({target} PROPERTIES {property} "{" ".join(values)}")\n'

    def target_link_options(self, target, values):
        if values:
            self.contents += f'target_link_options({target} PUBLIC "{" ".join(values)}")\n'

    def target_link_libraries(self, target, libraries):
        if libraries:
            self.contents += f'target_link_libraries({target} {" ".join(libraries)})\n'

    def link_libraries(self, libraries):
        if libraries:
            self.contents += f'link_libraries({" ".join(libraries)})\n'

    def target_include_directories(self, target, directories):
        if directories:
            self.contents += f'target_include_directories({target} PUBLIC {" ".join(sorted(directories))})\n'

    def add_custom_target(self, target, command, depends):
        self.contents += f'add_custom_target({target}\n' \
                         f'    COMMAND {command}\n' \
                         f'    DEPENDS {" ".join(depends)}\n' \
                         f')\n'

    def add_custom_command(self, target, when, command):
        self.contents += f'add_custom_command(\n' \
                         f'    TARGET {target}\n' \
                         f'    {when}\n' \
                         f'    COMMAND COMMAND {command}\n' \
                         f')\n'

    def set_compilers(self, toolchain_prefix):
        self.contents += f'set(CMAKE_C_COMPILER {toolchain_prefix}gcc)\n'
        self.contents += f'set(CMAKE_CXX_COMPILER {toolchain_prefix}g++)\n'
        self.contents += f'set(CMAKE_AR {toolchain_prefix}ar)\n'



import os

import docker
import subprocess
import signal
import sys

from cpm.infrastructure import filesystem
from cpm.domain import constants


class ProjectCommands(object):
    def build(self, project):
        if not filesystem.directory_exists(constants.BUILD_DIRECTORY):
            filesystem.create_directory(constants.BUILD_DIRECTORY)
        self.__build(project, [project.name], post_build=self.__post_build(project))

    def clean(self, project):
        if filesystem.directory_exists(constants.BUILD_DIRECTORY):
            filesystem.remove_directory(constants.BUILD_DIRECTORY)
        _ignore_exception(lambda: filesystem.delete_file(constants.CMAKELISTS))

    def build_tests(self, project, files_or_dirs):
        if not filesystem.directory_exists(constants.BUILD_DIRECTORY):
            filesystem.create_directory(constants.BUILD_DIRECTORY)
        if not files_or_dirs:
            self.__build_tests(project, ['tests'])
        else:
            tests_to_run = self.tests_from_args(project, files_or_dirs)
            self.__build_tests(project, [test.name for test in tests_to_run])

    def __build_tests(self, project, goals, post_build=''):
        if project.target.test_image and project.target.test_dockerfile:
            print('cpm: warning: both "test_image" and "test_dockerfile" options are specified, will use "test_image"')
        if project.target.test_image:
            self.__build_using_image(
                project,
                project.target.test_image,
                goals,
                post_build
            )
        elif project.target.test_dockerfile:
            self.__build_using_dockerfile(
                project,
                project.target.test_dockerfile,
                self.__test_image_name(project),
                goals,
                post_build
            )
        else:
            self.__build_goal(goals)

    def __build(self, project, goals, post_build=''):
        if project.target.image and project.target.dockerfile:
            print('cpm: warning: both "image" and "dockerfile" options are specified, will use "image"')
        if project.target.image:
            self.__build_using_image(
                project,
                project.target.image,
                goals,
                post_build
            )
        elif project.target.dockerfile:
            self.__build_using_dockerfile(
                project,
                project.target.dockerfile,
                self.__build_image_name(project),
                goals,
                post_build
            )
        else:
            self.__build_goal(goals)

    def __build_goal(self, goals):
        if any(result.returncode != 0 for result in [
            self.__run_command(constants.CMAKE_COMMAND, '-G', 'Ninja', '..'),
            self.__run_command(constants.NINJA_COMMAND, *goals)
        ]):
            raise BuildError

    def __build_using_image(self, project, image_name, goals, post_build):
        print(f'cpm: using Docker image {image_name}')
        client = docker.from_env()
        try:
            client.images.pull(image_name)
        except docker.errors.ImageNotFound:
            raise DockerImageNotFound(image_name)
        except docker.errors.NotFound:
            raise DockerImageNotFound(image_name)
        self.__build_inside_container(client, project, image_name, goals, post_build)

    def __build_using_dockerfile(self, project, dockerfile, image_name, goals, post_build):
        print(f'cpm: building image from {dockerfile}')
        client = docker.from_env()
        with open(dockerfile, 'rb') as fileobj:
            client.images.build(path='.', fileobj=fileobj, tag=image_name)
        self.__build_inside_container(client, project, image_name, goals, post_build)

    def __build_inside_container(self, client, project, image_name, goals, post_build):
        filesystem.create_file(
            f'{constants.BUILD_DIRECTORY}/build.sh',
            f'{constants.CMAKE_COMMAND} -G Ninja /{project.name} && {constants.NINJA_COMMAND} {" ".join(goals)}\n'
            f'{post_build}'
        )
        container = client.containers.run(
            image_name,
            command=f'sh /{project.name}/build/build.sh',
            working_dir=f'/{project.name}/build',
            volumes={f'{os.getcwd()}': {'bind': f'/{project.name}', 'mode': 'rw'}},
            user=f'{os.getuid()}:{os.getgid()}',
            environment=[f'PROJECT_NAME={project.name}', f'PROJECT_VERSION={project.version}'],
            detach=True
        )
        print(f'cpm: building inside {container.short_id}')
        for log in container.logs(stream=True):
            sys.stdout.write(log.decode())
        exit_code = container.wait()
        if exit_code['StatusCode'] != 0:
            raise BuildError
        container.remove()

    def __build_image_name(self, project):
        return f'{project.name}_cpm_build'.lower()

    def __test_image_name(self, project):
        return f'{project.name}_cpm_test'

    def run_tests(self, project, files_or_dirs, test_args=()):
        tests_to_run = project.test.test_suites if not files_or_dirs else self.tests_from_args(project, files_or_dirs)
        if project.target.test_image:
            test_results = self.__run_tests_using_image(
                project,
                project.target.test_image,
                tests_to_run,
                test_args
            )
        elif project.target.test_dockerfile:
            test_results = self.__run_tests_using_dockerfile(
                project,
                self.__test_image_name(project),
                tests_to_run,
                test_args
            )
        else:
            test_results = [self.run_test(test.name, test_args) for test in tests_to_run]
        if any(result != 0 for result in test_results):
            raise TestsFailed('tests failed')

    def run_test(self, executable, test_args):
        result = subprocess.run(
            [f'./{constants.BUILD_DIRECTORY}/{executable}'] + test_args
        )
        if result.returncode < 0:
            print(f'cpm: {executable} failed with {result.returncode} ({signal.Signals(-result.returncode).name})')
        return result.returncode

    def __run_tests_using_image(self, project, image_name, tests_to_run, test_args):
        client = docker.from_env()
        try:
            client.images.pull(image_name)
        except docker.errors.ImageNotFound:
            raise DockerImageNotFound(image_name)
        except docker.errors.NotFound:
            raise DockerImageNotFound(image_name)
        return self.__run_tests_inside_container(project, client, image_name, tests_to_run, test_args)

    def __run_tests_using_dockerfile(self, project, image_name, tests_to_run, test_args):
        client = docker.from_env()
        return self.__run_tests_inside_container(project, client, image_name, tests_to_run, test_args)

    def __run_tests_inside_container(self, project, client, image_name, tests_to_run, test_args):
        container = client.containers.run(
            image_name,
            command=f'/bin/bash',
            working_dir=f'/{project.name}',
            volumes={f'{os.getcwd()}': {'bind': f'/{project.name}', 'mode': 'rw'}},
            user=f'{os.getuid()}:{os.getgid()}',
            detach=True,
            stdin_open=True,
        )
        container.start()
        results = [self.__run_test_inside_container(project, container, test.name, test_args) for test in tests_to_run]
        container.stop()
        container.remove()
        return results

    def __run_test_inside_container(self, project, container, executable, test_args):
        (exit_code, output) = container.exec_run(
            f'./{constants.BUILD_DIRECTORY}/{executable} {" ".join(test_args)}',
            workdir=f'/{project.name}',
            user=f'{os.getuid()}:{os.getgid()}'
        )
        sys.stdout.write(output.decode())
        result = exit_code
        if result < 0:
            print(f'cpm: {executable} failed with {result} ({signal.Signals(-result).name})')
        return result

    def __run_command(self, *args, cwd=constants.BUILD_DIRECTORY):
        return subprocess.run([*args], cwd=cwd)

    def __post_build(self, project):
        return '\n'.join([f'( cd .. && {post_build} )' for post_build in project.target.post_build])

    def tests_from_args(self, project, files_or_dirs):
        if not files_or_dirs:
            return 'tests'
        tests = []
        for arg in files_or_dirs:
            if filesystem.is_file(arg):
                tests.append(self.test_from_test_file(project, arg))
            elif filesystem.is_directory(arg):
                tests.extend(self.test_list_from_directory(project, arg))
        return tests

    def test_list_from_directory(self, project, directory):
        return [self.test_from_test_file(project, test_file) for test_file in filesystem.find(directory, 'test_*.cpp')]

    def test_from_test_file(self, project, test_file):
        return next(test for test in project.test.test_suites if test.main == test_file)


def _ignore_exception(call):
    try:
        call()
    except:
        pass


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


class TestsFailed(RuntimeError):
    pass


class BuildError(RuntimeError):
    pass


class DockerImageNotFound(RuntimeError):
    def __init__(self, image_name):
        super(DockerImageNotFound, self).__init__()
        self.image_name = image_name



from cpm.domain.project import project_descriptor_parser, project_composer
from cpm.domain.project.project_descriptor_parser import ProjectDescriptorNotFound
from cpm.domain.constants import bit_directory


class ProjectLoader(object):
    def load(self, directory, target_name='default'):
        project_descriptor = project_descriptor_parser.parse_from(directory)
        if not target_is_valid(project_descriptor, target_name):
            raise InvalidTarget
        # TODO: Target specific bits
        self.parse_bit_build_descriptors(project_descriptor.build.bits,
                                         project_descriptor.build.declared_bits,
                                         lambda description: description.build.declared_bits)
        self.parse_bit_build_descriptors(project_descriptor.test.bits,
                                         project_descriptor.test.declared_bits,
                                         lambda description: description.test.declared_bits)
        return project_composer.compose(project_descriptor, target_name)

    def parse_bit_build_descriptors(self, bits, declared_bits, next_declared_bits):
        for declared_bit in declared_bits:
            if declared_bit.name not in bits:
                try:
                    bit_descriptor = project_descriptor_parser.parse_from(bit_directory(declared_bit.name, declared_bit.version))
                    bit_descriptor.declared_bit = declared_bit
                    bits[declared_bit.name] = bit_descriptor
                    self.parse_bit_build_descriptors(bits, next_declared_bits(bit_descriptor), next_declared_bits)
                except ProjectDescriptorNotFound:
                    print(f'cpm: warning: bit \'{declared_bit.name}:{declared_bit.version}\' not installed. '
                          f'Run \'cpm install\' to install missing bits.')


def target_is_valid(project_descriptor, target_name):
    return target_name == 'default' or any(target.name == target_name for target in project_descriptor.targets.values())


class InvalidTarget(RuntimeError):
    pass



from dataclasses import dataclass, field

from cpm.domain.project.project_descriptor import ProjectDescriptor


@dataclass
class Package:
    path: str
    sources: list = field(default_factory=list)
    cflags: list = field(default_factory=list)
    cppflags: list = field(default_factory=list)
    ldflags: list = field(default_factory=list)
    include_directories: set = field(default_factory=set)


@dataclass
class Target:
    name: str = 'default'
    executable: str = ''
    main: str = 'main.cpp'
    image: str = ''
    dockerfile: str = ''
    test_image: str = ''
    toolchain_prefix: str = ''
    post_build: list = field(default_factory=list)
    packages: list = field(default_factory=list)
    include_directories: set = field(default_factory=set)
    cflags: list = field(default_factory=list)
    cppflags: list = field(default_factory=list)
    ldflags: list = field(default_factory=list)
    libraries: list = field(default_factory=list)
    bits: list = field(default_factory=list)


@dataclass
class TestSuite:
    name: str
    main: str
    packages: list = field(default_factory=list)
    include_directories: set = field(default_factory=set)
    cflags: list = field(default_factory=list)
    cppflags: list = field(default_factory=list)
    ldflags: list = field(default_factory=list)
    libraries: list = field(default_factory=list)


@dataclass
class Test:
    test_suites: list = field(default_factory=list)
    packages: list = field(default_factory=list)
    include_directories: set = field(default_factory=set)
    cflags: list = field(default_factory=list)
    cppflags: list = field(default_factory=list)
    ldflags: list = field(default_factory=list)
    libraries: list = field(default_factory=list)
    bits: list = field(default_factory=list)


@dataclass
class Project:
    name: str = ''
    version: str = '0.1'
    description: str = ''
    descriptor: ProjectDescriptor = field(default_factory=ProjectDescriptor)
    target: Target = field(default_factory=Test)
    test: Test = field(default_factory=Test)
    declared_bits: list = field(default_factory=list)
    actions: list = field(default_factory=list)
